% .
% ---------------------------------------------------------------------
% |                                                                   |  
% |                           VISTALAB:                               |
% |         The Matlab Toolbox for Spatio-Temporal Vision Models      |
% |                                                                   |
% |                     (version 4.0, june 2018)                      |  
% |              (c) Universitat de Valencia. 1996-2018               |
% |                                                                   |
% |                Jesus Malo          jesus.malo@uv.es               |
% |                                                                   |
% |   Contributors to previous versions (known as BasicVideoTools):   |
% |                Juan Gutierrez (random-dots video sequences)       |                           
% |                Valero Laparra (beta tester, thanks!)              |
% |                                                                   |
% |   Citation:                                                       |
% |                                                                   |  
% |    1st paper: Malo, Gutirrez, Epifanio, Ferri                    | 
% |               Perceptually weighted optical flow for motion-based |
% |               segmentation in MPEG-4 paradigm                     |
% |               Electr. Lett. 36 (20):1693-1694 (2000)              |
% |                                                                   |
% |    Code & web site:                                               |
% |               Malo & Guatierrez                                   | 
% |               VISTALAB: the Matlab toolboox for spatio-temporal   |
% |               vision models. Univ. Valencia 2014                  |
% |               http://isp.uv.es/code/visioncolor/vistalab.html     |
% |                                                                   |
% |                                                                   |
% ---------------------------------------------------------------------
%
% VISTALAB is a Matlab toolbox with the spatio-temporal mechanisms 
% (receptive fields and filters) to build texture and motion vision models. 
% VISTALAB includes spatio-temporal sampling functions and tools to visualize 
% spatio-temporal Fourier transforms. This allows accurate control of
% spatio-frequency and perceptual properties of achromatic movies.
% 
% Spatio-temporal sampling and stimuli control
% --------------------------------------------
%    
%    spatio_temp_freq_domain - Define 3D-spatio-temporal and 3D-frequency domains to build analytical sequences therein
%    spatio_temp_DCT_domain  - Define 3D-spatio-temporal and 3D-DCT-frequency domains to build analytical sequences therein
%    control_lum_contrast    - Imposes certain average luminance and std-contrast in an achromatic natural image
%    combine_test_background - combines two controlled images in a center-surround setting 
%    image_sequence          - Sequences from still images moving at controlled speed
%    noise_sequence          - Sequences from colored noise of controlled speed
%    dots_sequence           - Sequences from a set of objects and a flow field 
%       radial_flow          - Computes flow for ego-motion parallel to optical axis
%       lateral_flow         - Computes flow for ego-motion orthogonal to optical axis
%       circular_flow        - Circular flow
%       sinusoidal_flow      - Motion grating 
%    newtonian_sequence      - Sequence from a rigid solid moving in force fields
%       elipso3              - Definition of facets and trapezoids of an ellipsoid
%       dinam_tr             - Translation and rotation newtonian dynamics (Runge-Kutta integration) 
%       pintael2             - Illumination of facets and projection onto the camera
%
% Spatio-temporal transforms
% --------------------------
%    
%    fft3               - Computation of spatio-temporal FFT
%    ifft3              - Inverse FFT
%    show_fft3          - Visualization of 3D FFT
%    show_project_fft3  - Visualization of integrals
%    temporal_window    - Convenient to reduce edge effects in spatio-temporal filtering
%    dct3               - Computation of spatio-temporal DCT
%    idct3              - Inverse of 3D DCT
%    show_dct3          - Visualization of 3D DCT
%
% Spatio-temporal windows of visibility (Contrast Sensitivity Funcions)
% ---------------------------------------------------------------------
%
%    spatio_temp_CSF    - Spatio-temporal CSF (D.H. Kelly JOSA 79)
%    csfsso             - Achromatic CSF of the Standard Spatial Observer (Watson & Malo IEEE ICIP 02)
%    csf_chrom          - Red-Green & Yellow-Blue chromatic CSFs (Mullen Vis. Res. 85)
%
% Spatio-temporal linear receptive fields in LGN, V1 and MT
% ---------------------------------------------------------
%
%    sens_lgn3d_space   - Linear LGN filters (Difference of Gaussians and derivatives of Gaussians, Cai et al. J. Neurophysiol. 97)
%    sens_gabor3d_space - Linear V1 filters (spatio-temporal Gabors, Simoncelli & Heeger Vis.Res. 98)
%    sens_gabor3d_freq  - Linear V1 filters (defined in the frequency domain)
%    sens_MT            - Linear MT filters (coherent combinations of 3D Gabors in the freq. domain, Simoncelli & Heeger Vis.Res. 98).
%
% Read / write video data: from files to 3D or 4D arrays (height*width*frames or height*width*color*frames)
% ------------------------
%
%    read_video   - convenient interface for VideoReader for straightforward avi-to-matlab conversion
%    read_vqeg_mg - Simplified function to read achromatic part of original VQEG movies    
%    read_live_mg - Simplified function to read achromatic part of original LIVE movies
%    yuv2mov   - read yuv format movies (e.g. LIVE) and convert them to Matlab movies
%    read_vqeg - read VQEG files (both original and distorted)
%    build_achrom_movie_avi -  from 3D and 2D arrays to Matlab movies and avi files
%    build_color_avi - from 4D arrays (spatio-chromatic-temporal) to avi files
%
% Rearrange video data: 
% ---------------------
%
%    im2colcube - From 3D arrays (or 3D patches) to vectors
%    col2imcube - From vectors to 3D arrays
%    now2then   - From 3D arrays (modern format) to 2D arrays (old film-like format)
%    then2now   - from 2D arrays to 3D arrays
%
% Movie visualization (achromatic only):
% --------------------------------------
%
%    build_achrom_movie -  from 3D and 2D arrays to Matlab movies
%    disp_spatio_temp_patches - Displays N^2 spatiotemporal patches at the same time
%    disp_patches - Displays N^2 spatial patches at the same time
%    disp_patches_norm - Displays N^2 spatial patches at the same time (controlled normalization)
%    implay (not a part of this toolbox): opens an interactive video player
%
% Motion estimation and compensation (separate toolbox)
% ----------------------------------
%
%  See:   VistaVideoCoding: perceptually weighted optical flow computation
%         http://isp.uv.es/code/visioncolor/Video_coding.html       
%         http://www.scholarpedia.org/article/Optic_flow
%
% Spatio-Chromatic vision model (separate toolbox)
% -----------------------------
%
%  See:   BioMultiLayer_L_NL_color: biologically plausible deep neural network
%         http://isp.uv.es/code/visioncolor/vistamodels.html
%
% Demos
% -----
%
%    demo_vistalab                - Demo on how to use the above functions (except random dots and newtonian sequences)
%    example_random_dots_sequence - Demo on random dots sequences with controlled flow
%    example_newtonian_sequence   - Demo on physics-controlled sequences
%
% Sample movies
% -------------
% 
%    real_seq_Nx_75_Nf_64.mat     - a natural sequence with illustrative motion: hands moving at about 1 deg/sec 
%                                   when fsx = 37.5 cpd and fst = 24 Hz
%                                   NOTE: it is in 2D array format (see then2now.m)
%

% Modifications from previous versions
%%%%%%%%%%%%%%%%%

%    csfsso 

%    csf_chrom

%    disp_patches

%    disp_spatio_temp_patches

%    sens_gabor3d_space

%    control_lum_comtrast

%    build_achrom_movie_avi

%    sens_MT

%    control_lum_contrast

%    noise_sequence

%    combine_test_background

%    spatial_csf_speed

%    dct3, idct3, spatio_temp_DCT_domain, show_dct3, demo_dct3d
